#include <proto.h>

#include "gtmatrix.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtMatrix class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/************************************************************************\
*
* Create and init matrix.
*
\************************************************************************/
GtMatrix *GtMkMatrix(char *n,GtManager *p,GtTag *tgs)
{
	GtMatrix *obj;
	
	if(!(obj=new GtMatrix()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtMatrix object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init matrix.
*
\************************************************************************/
GtMatrix *GtMkMatrix(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtMatrix *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkMatrix(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty matrix.
*
\************************************************************************/
GtMatrix::GtMatrix(void)
{
	rows=1;
	cols=1;
	rmin=NULL;
	rmax=NULL;
	cmin=NULL;
	cmax=NULL;
	orient=GTO_HORIZ;
	ftype=GTF_NONE;
	space=0;
	ibw=0;
	bdrw=1;
	ibdrw=0;
	unisize=FALSE;
}

/************************************************************************\
*
* Destroy matrix.
*
\************************************************************************/
void GtMatrix::Destroy(void)
{
	if(rmin)
		free(rmin);
	GtBdrMan::Destroy();
}

/************************************************************************\
*
* Setup matrix.
*
\************************************************************************/
sbool GtMatrix::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	
	if(!GtBdrMan::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_ORIENT,GTRT_ORIENT,&orient);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	app->GetRes(this,GTR_SPACE,GTRT_BYTE,&space);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	app->GetRes(this,GTR_UNISZ,GTRT_BOOL,&unisize);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Orient:
				orient=t->d.val;
				if(orient>GTF_LAST_NR)
					orient=GTO_HORIZ;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_NONE;
				break;
			case GT_Spacing:
				space=t->d.val;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
			case GT_Columns:
				if(t->d.val)
					cols=t->d.val;
				break;
			case GT_Rows:
				if(t->d.val)
					rows=t->d.val;
				break;
			case GT_UniSize:
				if(t->d.val)
					unisize=TRUE;
				else
					unisize=FALSE;
				break;
		}
	}
	maxchild=cols*rows;
	if(!(rmin=(word *)malloc(sizeof(word)*(4*cols+4*rows))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for size vectors!\n");
		return(FALSE);
	}
	rmax=rmin+rows;
	rsize=rmax+rows;
	ridx=rsize+rows;
	cmin=ridx+rows;
	cmax=cmin+cols;
	csize=cmax+cols;
	cidx=csize+cols;
	switch(ftype)
	{
		case GTF_NONE:
			ibw=ibdrw;
			break;
		case GTF_SIMPLE:
		case GTF_IN:
		case GTF_OUT:
			ibw=ibdrw+bdrw;
			break;
		case GTF_DBL_IN:
		case GTF_DBL_OUT:
			ibw=ibdrw+2*bdrw;
			break;
	}
	return(TRUE);
}

/************************************************************************\
*
* Set matrix attributes.
*
\************************************************************************/
sbool GtMatrix::Set(GtTag *tgs)
{
	GtTag *t;
	sbool ch=FALSE,vch=FALSE;
	uword nr=0,nc=0;
	
	if(!GtBdrMan::Set(tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Orient:
				orient=t->d.val;
				if(orient>GTF_LAST_NR)
					orient=GTO_HORIZ;
				ch=TRUE;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_NONE;
				vch=TRUE;
				break;
			case GT_Spacing:
				space=t->d.val;
				ch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Columns:
				if(t->d.val)
					nc=t->d.val;
				break;
			case GT_Rows:
				if(t->d.val)
					nr=t->d.val;
				break;
			case GT_UniSize:
				if(t->d.val)
					unisize=TRUE;
				else
					unisize=FALSE;
				ch=TRUE;
				break;
		}
	}
	if(nr || nc)
	{
		if(!nr)
			nr=rows;
		if(!nc)
			nc=cols;
		if(nrchild>nc*nr)
		{
			GtError(this,GTEC_MISC,"Unable to change rows/columns!\n");
			return(FALSE);
		}
		cols=nc;
		rows=nr;
		maxchild=cols*rows;
		if(rmin)
			free(rmin);
		if(!(rmin=(word *)malloc(sizeof(word)*(4*cols+4*rows))))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for size vectors!\n");
			return(FALSE);
		}
		rmax=rmin+rows;
		rsize=rmax+rows;
		ridx=rsize+rows;
		cmin=ridx+rows;
		cmax=cmin+cols;
		csize=cmax+cols;
		cidx=csize+cols;
		ch=TRUE;
	}
	if(ch || vch)
	{
		switch(ftype)
		{
			case GTF_NONE:
				ibw=ibdrw;
				break;
			case GTF_SIMPLE:
			case GTF_IN:
			case GTF_OUT:
				ibw=ibdrw+bdrw;
				break;
			case GTF_DBL_IN:
			case GTF_DBL_OUT:
				ibw=ibdrw+2*bdrw;
				break;
		}
		this->Relayout();
		if(vch)
			this->Refresh();
	}
	return(TRUE);
}

/************************************************************************\
*
* Get matrix attributes.
*
\************************************************************************/
void GtMatrix::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtBdrMan::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Orient:
				*ptr=orient;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
			case GT_Spacing:
				*ptr=space;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
			case GT_Columns:
				*ptr=cols;
				break;
			case GT_Rows:
				*ptr=rows;
				break;
			case GT_UniSize:
				*ptr=unisize;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtMatrix::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtMatrix")==0)
		return(TRUE);
	else
		return(GtBdrMan::IsSubClassOf(cn));
}

/************************************************************************\
*
* Set matrix event mask.
*
\************************************************************************/
void GtMatrix::SetEMask(ulong em)
{
	GtBdrMan::SetEMask(em|ExposureMask);
}

/************************************************************************\
*
* Draw matrix.
*
\************************************************************************/
sbool GtMatrix::Draw(XEvent *xe)
{
	int n,xp,yp;
	
	if(xe && ((xe->type!=Expose) || (xe->xexpose.count!=0)))
		return(TRUE);
	if(flags & GTWF_NOWIN)
	{
		xp=xpos;
		yp=ypos;
	}
	else
	{
		xp=0;
		yp=0;
	}
	switch(ftype)
	{
		case GTF_SIMPLE:
			for(n=0; n<bdrw; n++)
				gtfore->DrawRectangle(gtwin,xp+n,yp+n,xsize-2*n-1,ysize-2*n-1,FALSE);
			break;
		case GTF_DBL_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtlht,gtdrk,gtwin,xp+bdrw+n,yp+bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_DBL_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtdrk,gtlht,gtwin,xp+bdrw+n,yp+bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
			}
			break;
	}
	return(TRUE);
}

/************************************************************************\
*
* Relayout matrix.
*
\************************************************************************/
void GtMatrix::Relayout(void)
{
	GtWidget *obj;
	GtWListBdr *li;
	int xs,ys;
	int mxs,mys,mxxs=GTS_MAX,mxys=GTS_MAX;
	int i,j,n,t;

	if(flags & GTWF_UNBUILT)
		return;
	mxs=2*ibw;
	mys=2*ibw;
	// Set alla tables to start values for calculations
	for(i=0; i<rows; i++)
	{
		rmin[i]=1;
		rmax[i]=GTS_MAX;
		ridx[i]=i;
	}
	for(j=0; j<cols; j++)
	{
		cmin[j]=1;
		cmax[j]=GTS_MAX;
		cidx[j]=j;
	}
	i=0;
	j=0;
	// Loop through all managed childrean
	for(li=(GtWListBdr *)GtSFirst(&man); li;
			li=(GtWListBdr *)GtSNext(&li->wl.lnk))
	{
		obj=li->wl.w;
		// Skip unmapped child but let it take one position
		if(obj->IsMapped())
		{
			// Calculate minimum size of child and it's column and row
			obj->GetMinSize(&xs,&ys);
			li->minxs=GTMAX(xs,li->minxs);
			li->minys=GTMAX(ys,li->minys);
			rmin[i]=GTMAX(rmin[i],li->minys+li->btop+li->bbot);
			cmin[j]=GTMAX(cmin[j],li->minxs+li->bleft+li->bright);
			// Calculate maximum size of child and it's column and row
			obj->GetMaxSize(&xs,&ys);
			li->maxxs=GTMIN(xs,li->maxxs);
			li->maxys=GTMIN(ys,li->maxys);
			rmax[i]=GTMIN(rmax[i],li->maxys+li->btop+li->bbot);
			cmax[j]=GTMIN(cmax[j],li->maxxs+li->bleft+li->bright);
			rmax[i]=GTMAX(rmin[i],rmax[i]);
			cmax[j]=GTMAX(cmin[j],cmax[j]);
			// Insert child in size-sorted rows list
			for(n=0; n<i; n++)
			{
				// Sort by ascending max min difference
				if(rmax[i]-rmin[i]<rmax[ridx[n]]-rmin[ridx[n]])
				{
					for(t=i; t>n; t--)
					{
						ridx[t]=ridx[t-1];
					}
					ridx[n]=i;
					break;
				}
			}
			// Insert child in size-sorted columns list
			for(n=0; n<j; n++)
			{
				// Sort by ascending max min difference
				if(cmax[j]-cmin[j]<cmax[cidx[n]]-cmin[cidx[n]])
				{
					for(t=j; t>n; t--)
					{
						cidx[t]=cidx[t-1];
					}
					cidx[n]=j;
					break;
				}
			}
		}
		// Step to next position according to orientation
		if(orient==GTO_HORIZ)
		{
			j++;
			if(j==cols)
			{
				j=0;
				i++;
			}
		}
		else
		{
			i++;
			if(i==rows)
			{
				i=0;
				j++;
			}
		}
	}
	// Are childrean to be uniformly sized
	if(unisize)
	{
		// Find smallest and biggest row
		for(i=0; i<rows; i++)
		{
			mys=GTMAX(mys,rmin[i]);
			mxys=GTMIN(mxys,rmax[i]);
		}
		// Find smallest and biggest column
		for(j=0; j<cols; j++)
		{
			mxs=GTMAX(mxs,cmin[j]);
			mxxs=GTMIN(mxxs,cmax[j]);
		}
		mys=GTMAX(mys,1);
		mxs=GTMAX(mxs,1);
		mxys=GTMAX(mys,mxys);
		mxxs=GTMAX(mxs,mxxs);
		// Set all rows and all columns to the same sizes
		for(i=0; i<rows; i++)
		{
			rmin[i]=mys;
			rmax[i]=mxys;
		}
		for(j=0; j<cols; j++)
		{
			cmin[j]=mxs;
			cmax[j]=mxxs;
		}
		this->SetMinMaxSize(2*ibw+(mxs+space)*cols-space,
												2*ibw+(mys+space)*rows-space,
												2*ibw+(mxxs+space)*cols-space,
												2*ibw+(mxys+space)*rows-space);
	}
	else
	{
		// Calculate size of matrix by adding sizes of rows and columns
		for(i=0; i<rows; i++)
		{
			mys+=rmin[i]+space;
			mxys+=rmax[i]+space;
			if(mxys>GTS_MAX)
				mxys=GTS_MAX;
		}
		for(j=0; j<cols; j++)
		{
			mxs+=cmin[j]+space;
			mxxs+=cmax[j]+space;
			if(mxxs>GTS_MAX)
				mxxs=GTS_MAX;
		}
		mys=GTMAX(mys,2*ibw+1);
		mxs=GTMAX(mxs,2*ibw+1);
		mxys=GTMAX(mys,mxys);
		mxxs=GTMAX(mxs,mxxs);
		this->SetMinMaxSize(mxs-space,mys-space,mxxs-space,mxys-space);
	}
}

/************************************************************************\
*
* Layout matrix.
*
\************************************************************************/
void GtMatrix::Layout(void)
{
	GtWidget *obj;
	GtWListBdr *li;
	int xp,yp,xs,ys,nxs,nys,i,j,ctw,rtw,tsz,xpe,ype;

	if(flags & GTWF_UNBUILT)
		return;
	if(xsize>minxs)
		xs=xsize-minxs;
	else
		xs=0;
	if(ysize>minys)
		ys=ysize-minys;
	else
		ys=0;
	rtw=rows;
	for(i=0; i<rows; i++)
	{
		tsz=ys/rtw;
		tsz=GTMIN(tsz,rmax[ridx[i]]-rmin[ridx[i]]);
		rsize[ridx[i]]=tsz+rmin[ridx[i]];
		ys-=tsz;
		rtw--;
	}
	ctw=cols;
	for(j=0; j<cols; j++)
	{
		tsz=xs/ctw;
		tsz=GTMIN(tsz,cmax[cidx[j]]-cmin[cidx[j]]);
		csize[cidx[j]]=tsz+cmin[cidx[j]];
		xs-=tsz;
		ctw--;
	}
	i=0;
	j=0;
	xp=ibw;
	yp=ibw;
	if(flags & GTWF_NOWIN)
	{
		xpe=xpos;
		ype=ypos;
	}
	else
	{
		xpe=0;
		ype=0;
	}
	for(li=(GtWListBdr *)GtSFirst(&man); li;
			li=(GtWListBdr *)GtSNext(&li->wl.lnk))
	{
		obj=li->wl.w;
		nxs=csize[j];
		nys=rsize[i];
		obj->SetPosSize(xpe+xp+li->bleft,ype+yp+li->btop,
				nxs-li->bleft-li->bright,nys-li->btop-li->bbot);
		if(orient==GTO_HORIZ)
		{
			j++;
			xp+=nxs+space;
			if(j==cols)
			{
				j=0;
				xp=ibw;
				i++;
				yp+=nys+space;
			}
		}
		else
		{
			i++;
			yp+=nys+space;
			if(i==rows)
			{
				i=0;
				yp=ibw;
				j++;
				xp+=nxs+space;
			}
		}
	}
}
